/*
 *
 *
 * 
 * http://www.suse.de/~mana/alsa090_howto.html#sect02
 *
 */

#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <alsa/asoundlib.h>

#include "generator.h"

#define GENERATOR_MAX_THREADS 16

struct s_generator_cfg {
	/* generator */
	char *device;
	int freq;
	pthread_t thread;
	int done;
	int running;

	/* alsa */
	snd_pcm_t *pcm_handle;
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_uframes_t periodsize;
	int num_frames,
	    exact_rate;
	
} generator_cfg[GENERATOR_MAX_THREADS];

static void *generator_main(void *arg);

/*****************************************************************************
 * 	ALSA
 ****************************************************************************/

int genalsa_init(struct s_generator_cfg *cfg) {
	int rate = 44100;	/* Sample rate */
	int exact_rate = 0;	/* Sample rate returned by
				 * snd_pcm_hw_params_set_rate_near */ 
	int dir = 0;		/* exact_rate == rate --> dir = 0
				 * exact_rate < rate  --> dir = -1
				 * exact_rate > rate  --> dir = 1 */
	int periods = 2;	/* Number of periods */
	
	cfg->periodsize = 8192; /* Periodsize (bytes) */
	snd_pcm_stream_t stream = SND_PCM_STREAM_PLAYBACK;

	/* Allocate the snd_pcm_hw_params_t structure on the stack. */
	snd_pcm_hw_params_alloca(&(cfg->hwparams));
	
	/* open device */
	if (snd_pcm_open(&(cfg->pcm_handle), cfg->device, stream, 0) < 0) {
		fprintf(stderr, "Error opening PCM device %s\n", cfg->device);
		return(-1);
	}

	/* Init hwparams with full configuration space */
	if (snd_pcm_hw_params_any(cfg->pcm_handle, cfg->hwparams) < 0) {
		fprintf(stderr, "Can not configure this PCM device.\n");
		return(-1);
	}

	/* Set access type. This can be either    */
	/* SND_PCM_ACCESS_RW_INTERLEAVED or       */
	/* SND_PCM_ACCESS_RW_NONINTERLEAVED.      */
	/* There are also access types for MMAPed */
	/* access, but this is beyond the scope   */
	/* of this introduction.                  */
	if (snd_pcm_hw_params_set_access(cfg->pcm_handle, cfg->hwparams, SND_PCM_ACCESS_RW_INTERLEAVED) < 0) {
		fprintf(stderr, "Error setting access.\n");
		return(-1);
	}

	/* Set sample format */
	if (snd_pcm_hw_params_set_format(cfg->pcm_handle, cfg->hwparams, SND_PCM_FORMAT_S16_LE) < 0) {
		fprintf(stderr, "Error setting format.\n");
		return(-1);
	}

	/* Set sample rate. If the exact rate is not supported */
	/* by the hardware, use nearest possible rate.         */ 
	cfg->exact_rate = rate;
	if (snd_pcm_hw_params_set_rate_near(cfg->pcm_handle, cfg->hwparams, (unsigned *) &(cfg->exact_rate), 0) < 0) {
		fprintf(stderr, "Error setting rate.\n");
		return(-1);
	}
	if (rate != cfg->exact_rate) {
		fprintf(stderr, "The rate %d Hz is not supported by your hardware.\n"
				"==> Using %d Hz instead.\n", rate, cfg->exact_rate);
	}

	/* Set number of channels */
	if (snd_pcm_hw_params_set_channels(cfg->pcm_handle, cfg->hwparams, 2) < 0) {
		fprintf(stderr, "Error setting channels.\n");
		return(-1);
	}

	/* Set number of periods. Periods used to be called fragments. */ 
	if (snd_pcm_hw_params_set_periods(cfg->pcm_handle, cfg->hwparams, periods, 0) < 0) {
		fprintf(stderr, "Error setting periods.\n");
		return(-1);
	}

	/* Set buffer size (in frames). The resulting latency is given by */
	/* latency = periodsize * periods / (rate * bytes_per_frame)     */
	if (snd_pcm_hw_params_set_buffer_size(cfg->pcm_handle, cfg->hwparams, (cfg->periodsize * periods)>>2) < 0) {
		fprintf(stderr, "Error setting buffersize.\n");
		return(-1);
	}

	/* Apply HW parameter settings to */
	/* PCM device and prepare device  */
	if (snd_pcm_hw_params(cfg->pcm_handle, cfg->hwparams) < 0) {
		fprintf(stderr, "Error setting HW params.\n");
		return(-1);
	}

	cfg->num_frames = cfg->periodsize >> 2;
}

int genalsa_play(struct s_generator_cfg *cfg, short *buffer) {

	/* Write num_frames frames from buffer data to    */ 
	/* the PCM device pointed to by pcm_handle.       */
	/* Returns the number of frames actually written. */
	/* snd_pcm_sframes_t snd_pcm_writei(cfg->pcm_handle, data, num_frames); */

	/* Write num_frames frames from buffer data to    */ 
	/* the PCM device pointed to by pcm_handle.       */ 
	/* Returns the number of frames actually written. */
	/* snd_pcm_sframes_t snd_pcm_writen(cfg->pcm_handle, data, num_frames); */

	unsigned char *data;
	int pcmreturn, i;
	short s1, s2;
	int frames, w;

	frames = cfg->num_frames;
	data = (unsigned char *)malloc(frames * 4);
	bzero(data, frames * 4);
	
	for(i = 0; i < frames; i++) {
		s1 = buffer[i];
		s2 = s1;
		data[4*i] = (unsigned char)s1;
		data[4*i+1] = s1 >> 8;
		data[4*i+2] = (unsigned char)s2;
		data[4*i+3] = s2 >> 8;
	}
	
#if 0
	w = 0;
	while ((w = snd_pcm_writei(cfg->pcm_handle, data + w * 4, frames - w)) < 0) {
		snd_pcm_prepare(cfg->pcm_handle);
		fprintf(stderr, "<<<<<<<<<<<<<<< Buffer Underrun >>>>>>>>>>>>>>>\n");
	}
#else
	while ((pcmreturn = snd_pcm_writei(cfg->pcm_handle, data, frames)) < 0) {
		snd_pcm_prepare(cfg->pcm_handle);
		fprintf(stderr, "<<<<<<<<<<<<<<< Buffer Underrun >>>>>>>>>>>>>>>\n");
	}
#endif
	free(data);
}

int genalsa_close(struct s_generator_cfg *cfg) {
	/* Stop PCM device and drop pending frames */
	snd_pcm_drop(cfg->pcm_handle);

	/* Stop PCM device after pending frames have been played */ 
	snd_pcm_drain(cfg->pcm_handle);
}


/*****************************************************************************
 * 	Generator API
 ****************************************************************************/

void generator_init(void) {
	int i;
	bzero(generator_cfg, sizeof(struct s_generator_cfg) * GENERATOR_MAX_THREADS);
}

void generator_set_freq(int thread, int freq) {
	generator_cfg[thread].freq = freq;
}

/** Start Generator
 *
 *	RETURN VALUE
 *	       On success, 0 is returned. On error, a non-zero error code is
 *	       returned.
 */
int generator_start(int thread, char *device, int freq) {
	if (!device)
		device = "default";

	if (!generator_cfg[thread].running) {
		generator_cfg[thread].running = 1;
		generator_cfg[thread].device = strdup(device);
		generator_cfg[thread].freq = freq;
		generator_cfg[thread].done = 0;
		return pthread_create(&(generator_cfg[thread].thread),
				NULL, generator_main,
				&(generator_cfg[thread]));
	} else {
		return -1;
	}
}

/** Stop generator
 * 
 *	RETURN VALUE
 *	       On success,  0 is returned.  On error, a non-zero error code is
 *	       returned.
 */
int generator_stop(int thread) {
	if (generator_cfg[thread].running) {
		generator_cfg[thread].done = 1;
		return pthread_join(generator_cfg[thread].thread, NULL);
	} else {
		return -1;
	}
}

/* Generator's thread
 */
void *generator_main(void *arg) {
	struct s_generator_cfg *cfg = (struct s_generator_cfg *)arg;
	short *buffer = NULL;
	int buf_size = 0;
	int i = 0, ampl = 0, minus = 0;
	long t = 0;
	
	

	ampl = (1 << (sizeof(short) * 8 - 3));
	buf_size = cfg->num_frames;
	buffer = (short*)malloc(buf_size * sizeof(short));
	
	//printf("Start: f = %d, ampl = %d\n", cfg->freq, ampl);
	
	/* ALSA Init */
	genalsa_init(cfg);

	while (!cfg->done) {
		/* calculate data */
		for (i = 0; i < buf_size; i++) {
			/*
			buffer[i] = (short)(sin((float)(t * 3.141592 * cfg->freq)
						/ (float)cfg->exact_rate * 2) * ampl);
			*/
			buffer[i] = (short)(sin((t * (double)cfg->freq)
						* 2.0 * 3.141592	/* na radiany */
						/ (double)cfg->exact_rate
						) * ampl);

			t++;
			
			/* v nule */
			if (minus) {
				if (buffer[i] == 0)
					t = 1;
				if (buffer[i] >= 0)
					minus = 0;
			} else if (buffer[i] < - (ampl / 4)) {
				minus = 1;
			}
			

		}
		
		/* play data */
		genalsa_play(cfg, buffer);
		
		/* little sleep */
		usleep(100);
	}

	free(buffer);
	
	/* ALSA clean up */
	genalsa_close(cfg);
	
	//printf("Stop:  f = %d\n", cfg->freq);
	
	free(cfg->device);
	//bzero(cfg, sizeof(struct s_generator_cfg));
	cfg->done = 1;
	cfg->running = 0;
}


